
local modmod = ModMod
local saved, dialog

local MAX_MODMODS = 10
local CLICK_FORMAT = "CLICK ModModKey%d:LeftButton"

SlashCmdList["MODMOD"] = function()
	modmod:SetShown(not modmod:IsVisible())
end
SLASH_MODMOD1 = "/modmod"

function modmod:OnEvent(event,...)
	if modmod[event] then
		modmod[event](self,...)
	end
end

function modmod:PLAYER_LOGIN()
	ModModSaved_Global = ModModSaved_Global or {{}}
	ModModSaved_PerCharacter = ModModSaved_PerCharacter or {{}}
	saved = GetCurrentBindingSet()==1 and ModModSaved_Global or ModModSaved_PerCharacter

	modmod.version:SetText(format("version %s",GetAddOnMetadata("ModMod","Version")))

	modmod.selectedModMod = 1
	modmod:RegisterEvent("UPDATE_BINDINGS")
	modmod:RegisterEvent("PLAYER_REGEN_DISABLED")
	modmod:RegisterEvent("ADDON_LOADED")
	modmod:UpdateSecureStuff()
end

function modmod:UPDATE_BINDINGS()
	saved = GetCurrentBindingSet()==1 and ModModSaved_Global or ModModSaved_PerCharacter
	if not InCombatLockdown() then
		modmod:UpdateSecureStuff() -- we could have changed binding sets
		if modmod:IsVisible() then
			modmod:UpdateUI()
		end
	else -- in combat, update secure stuff when leaving combat
		modmod:RegisterEvent("PLAYER_REGEN_ENABLED")
	end
end

-- can't have the window up in combat
function modmod:PLAYER_REGEN_DISABLED()
	modmod:Hide()
end

function modmod:PLAYER_REGEN_ENABLED()
	modmod:UnregisterEvent("PLAYER_REGEN_ENABLED")
	modmod:UpdateSecureStuff()
end

-- add ModMod button to the default binding UI when it's loaded
function modmod:ADDON_LOADED(addon)
	if addon=="Blizzard_BindingUI" then
		modmod:UnregisterEvent("ADDON_LOADED")
		local button = CreateFrame("Button",nil,KeyBindingFrame,"UIPanelButtonTemplate")
		button:SetSize(120,22)
		button:SetText("ModMod")
		button:SetPoint("RIGHT",KeyBindingFrame.unbindButton,"LEFT")
		button:SetScript("OnClick",function() -- click of ModMod button keybinding interface
			if not KeyBindingFrame.characterSpecificButton:GetChecked() and GetCurrentBindingSet()==CHARACTER_BINDINGS then
				CONFIRMED_DELETING_CHARACTER_SPECIFIC_BINDINGS = 1
			end
			OkayButton_OnClick(KeyBindingFrame)
			HideUIPanel(GameMenuFrame)
			modmod:UPDATE_BINDINGS()
			modmod:Show()
		end)
	end
end

function modmod:OnShow()
	if InCombatLockdown() then
		print("\124cffff2222The ModMod window can't be used in combat, sorry!")
		modmod:Hide()
		return
	end
	if not modmod.shown then
		dialog = modmod.dialog
		dialog.inset.bind:RegisterForClicks("AnyUp")
		modmod:CreateCategories()
		modmod:CreateActions()
		modmod.shown = true
	end
	modmod:UpdateUI()
	if not tContains(UISpecialFrames,"ModMod") then
		tinsert(UISpecialFrames,"ModMod")
	end
end

function modmod:OnHide()
	PlaySound(PlaySoundKitID and "gsTitleOptionExit" or SOUNDKIT.GS_TITLE_OPTION_EXIT)
	for i=#UISpecialFrames,1,-1 do
		if UISpecialFrames[i]=="ModMod" then
			tremove(UISpecialFrames,i)
		end
	end
end

function modmod:UpdateUI()
	saved = GetCurrentBindingSet()==1 and ModModSaved_Global or ModModSaved_PerCharacter
	local numModMods = #saved -- number of modmod keys saved
	if modmod.selectedModMod > numModMods then
		modmod.selectedModMod = numModMods
	end

	modmod.delete:SetEnabled(numModMods>1 or next(saved[1]) or GetBindingKey(format(CLICK_FORMAT,1)))
	modmod.new:SetEnabled(numModMods<MAX_MODMODS)

	modmod:UpdateModModsList()
	modmod:UpdateActionsData()
	modmod:UpdateCategoriesData()
	modmod:UpdateCategoriesList()
	modmod:UpdateActionsList()
end

function modmod:GetNumBindsForModMod(index)
	local count = 0
	if saved[index] then
		for k,v in pairs(saved[index]) do
			count = count + 1
		end
	end
	return count
end

--[[ Keys: This is the left column, the modmod key list. ]]

-- updates the list of modmod keys in the left column
function modmod:UpdateModModsList()
	for i=1,#saved do
		local button = modmod.modModsList.buttons[i]
		button:SetPoint("TOPLEFT",6,-6-(i-1)*44)
		local texture = button:GetNormalTexture()
		if modmod.selectedModMod==i then
			texture:SetTexCoord(0.00390625,0.87890625,0.59179688,0.66992188)
		else
			texture:SetTexCoord(0.00390625,0.87890625,0.75195313,0.83007813)
		end
		local key = GetBindingKey(format(CLICK_FORMAT,i))
		button.key:SetText((not key or key=="") and "\124cff999999"..NOT_BOUND or GetBindingText(key))
		button.count:SetText(format("Bindings: %d",modmod:GetNumBindsForModMod(i)))
		button:Show()
	end
	for i=#saved+1,MAX_MODMODS do
		modmod.modModsList.buttons[i]:Hide()
	end
end

function modmod:KeysOnClick()
	PlaySound(PlaySoundKitID and "igMainMenuOptionCheckBoxOn" or SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON)
	modmod.selectedModMod = self:GetID()
	modmod.selectedCategory = 1
	modmod:UpdateUI()
end

--[[ Categories: This is the middle column, the binding categories. ]]

modmod.selectedCategory = 1
modmod.categories = {} -- list of categories (build only once)
modmod.categoriesCounts = {} -- number of actions in this category (UpdateCategoriesData)

-- modmod.categories is gathered only once (when UI first shown) and never changes.
function modmod:CreateCategories()
	for i=1,GetNumBindings() do
		local _,category = GetBinding(i)
		if category and not tContains(modmod.categories,category) then
			tinsert(modmod.categories,category)
		end
	end
	tinsert(modmod.categories,"OTHER")
	local scrollFrame = modmod.categoriesList.scrollFrame
	scrollFrame.stepSize = 300
	scrollFrame.update = modmod.UpdateCategoriesList
	HybridScrollFrame_CreateButtons(scrollFrame, "ModModCategoriesTemplate")
end

-- UpdateActionsData needs to run before this to populate actionKeyLookup
function modmod:UpdateCategoriesData()
	wipe(modmod.categoriesCounts)
	local counts = modmod.categoriesCounts
	for i=1,GetNumBindings() do
		local action,category = GetBinding(i)
		category = category or "OTHER"
		if modmod.actionKeyLookup[action] then
			counts[category] = (counts[category] or 0)+1
		end
	end
end

function modmod:UpdateCategoriesList()
	local numData = #modmod.categories
	local scrollFrame = modmod.categoriesList.scrollFrame
	local offset = HybridScrollFrame_GetOffset(scrollFrame)
	local buttons = scrollFrame.buttons
	for i=1,#buttons do
		local index = i + offset
		local button = buttons[i]
		button.index = index
		if ( index <= numData ) then
			button:SetID(index)
			local category = modmod.categories[index]
			local count = modmod.categoriesCounts[category]
			button:SetText(format(count and "%s \124cffffffff(%d)" or "%s",_G[category] or category,count))
			button:Show()
			local highlight = button:GetHighlightTexture()
			if index==modmod.selectedCategory then
				highlight:SetVertexColor(1,.82,0)
				button:LockHighlight()
			else
				highlight:SetVertexColor(.25,.5,1)
				button:UnlockHighlight()
			end
		else
			button:Hide()
		end
	end
	HybridScrollFrame_Update(scrollFrame, 20*numData, 20)
	-- adjust width of buttons based on scrollBar being visible or not
	local buttonWidth = scrollFrame.scrollBar:IsVisible() and 156 or 170
	for i=1,#buttons do
		buttons[i]:SetWidth(buttonWidth)
	end
end

function modmod:CategoryOnClick()
	PlaySound(PlaySoundKitID and "igMainMenuOptionCheckBoxOn" or SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON)
	modmod.selectedCategory = self.index
	modmod:UpdateUI()
end

--[[ Actions: This is the right-most column, the actions and their modmod binding. ]]

modmod.actions = {}
modmod.actionKeyLookup = {}

-- this just sets up the scrollFrame for the actions, runs only once.
function modmod:CreateActions()
	local scrollFrame = modmod.actionsList.scrollFrame
	scrollFrame.stepSize = 300
	scrollFrame.update = modmod.UpdateActionsList
	HybridScrollFrame_CreateButtons(scrollFrame, "ModModActionsTemplate") -- ,8,-8)
end

-- fills modmod.actions with the binding actions (ie "ACTIONBUTTON1") for selected category
function modmod:UpdateActionsData()
	wipe(modmod.actions)
	local targetCategory = modmod.categories[modmod.selectedCategory]
	for i=1,GetNumBindings() do
		local action,category = GetBinding(i)
		if ((not category and targetCategory=="OTHER") or category==targetCategory) then
			tinsert(modmod.actions,action)
		end
	end
	wipe(modmod.actionKeyLookup)
	for key,action in pairs(saved[modmod.selectedModMod]) do
		modmod.actionKeyLookup[action] = key
	end
end

function modmod:UpdateActionsList()
	local numData = #modmod.actions
	local scrollFrame = modmod.actionsList.scrollFrame
	local offset = HybridScrollFrame_GetOffset(scrollFrame)
	local buttons = scrollFrame.buttons
	for i=1,#buttons do
		local index = i + offset
		local button = buttons[i]
		button.index = index
		if ( index <= numData ) then
			local action = modmod.actions[index]
			if action:match("^HEADER") then
				button.header:Show()
				button.header:SetText(_G["BINDING_"..action] or (action:match("^HEADER_BLANK") and "") or (action:gsub("HEADER_","")))
				button.description:Hide()
				button.binding:Hide()
			else
				button.description:SetText(_G["BINDING_NAME_"..action] or action)
				button.header:Hide()
				button.description:Show()
				button.binding.action = action
				local key = modmod.actionKeyLookup[action]
				if key then
					button.binding:SetText(GetBindingText(key))
				else
					button.binding:SetText("\124cff666666"..NOT_BOUND)
				end
				button.binding:Show()
			end
			button:Show()
		else
			button:Hide()
		end
	end
	HybridScrollFrame_Update(scrollFrame, 24*numData, 24)
end

--[[ Panel buttons ]]

function modmod:PanelNewOrEditOnClick(edit)
	dialog.purpose = edit and "edit" or "new"
	dialog.title:SetText("Binding a ModMod key")
	dialog.defaultMessage = "This will be used as a modifier for other keys."
	modmod:DialogSetBind(edit and GetBindingKey(format(CLICK_FORMAT,modmod.selectedModMod)) or nil)
	PlaySound(PlaySoundKitID and "igMainMenuOptionCheckBoxOn" or SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON)
	dialog:Show()
end

function modmod:DeleteModMod(index)
	local numModMods = #saved
	modmod:DeleteBindsForModModIndex(index) -- remove key binding to the modmod
	if numModMods>1 then -- at least one modmod must be listed at all times
		modmod:DeleteBindsForModModIndex(index) -- just removes key binding to the modmod
		for i=index+1,numModMods do
			local key = GetBindingKey(format(CLICK_FORMAT,i))
			if key then
				SetBinding(key,format(CLICK_FORMAT,i-1))
			end
		end
		modmod:DeleteBindsForModModIndex(numModMods)
		tremove(saved,index)
		SaveBindings(GetCurrentBindingSet())
	else -- this is last modmod, key binding is already removed, now remove actions
		saved[index]={}
	end
	modmod:UpdateUI()
end

function modmod:PanelDeleteOnClick()
	local key = GetBindingKey(format(CLICK_FORMAT,modmod.selectedModMod))
	local numBinds = modmod:GetNumBindsForModMod(modmod.selectedModMod)
	if not key and numBinds==0 then -- don't prompt to confirm if unbound modmod has no binds
		modmod:DeleteModMod(modmod.selectedModMod)
	else -- prompt to confirm if modmod is bound or it has binds
		StaticPopupDialogs["MODMODDELETE"] = {
			text="Are you sure want to delete this ModMod"..(numBinds>0 and " and its bindings?" or "?"),
			button1=YES, button2=NO, timeout=0, whileDead=1,
			OnAccept=function() modmod:DeleteModMod(modmod.selectedModMod) end
		}
		StaticPopup_Show("MODMODDELETE")
	end
end

-- when clicking one of the action binding buttons in the right column
function modmod:PanelActionOnClick()
	dialog.purpose = "action"
	dialog.action = self.action
	dialog.title:SetText("Binding a ModMod action")
	local modModKey = GetBindingKey(format(CLICK_FORMAT,modmod.selectedModMod))
	modModKey = (not modModKey or modModKey=="") and "the ModMod key" or "\124cffffffff"..GetBindingText(modModKey).."\124r"
	dialog.defaultMessage = format("While %s is down, the chosen key will \124cffffffff%s\124r.",modModKey,_G["BINDING_NAME_"..self.action] or self.action)
	local existingKey
	for savedKey,savedAction in pairs(saved[modmod.selectedModMod]) do
		if savedAction==self.action then
			existingKey = savedKey
		end
	end
	modmod:DialogSetBind(existingKey)
	PlaySound(PlaySoundKitID and "igMainMenuOptionCheckBoxOn" or SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON)
	dialog:Show()
end

--[[ Bind Dialog ]]

function modmod:DialogSetBind(bind)
	dialog.inset.message:SetText(dialog.defaultMessage)
	local button = dialog.inset.bind
	button:SetNormalFontObject(bind and "GameFontHighlight" or "GameFontDisable")
	button:SetText(bind and GetBindingText(bind) or NOT_BOUND)
	button.binding = bind
	dialog.unbind:SetEnabled(bind and true)
end

function modmod:DialogBindOnKeyDown(key)
	-- modifier keys are never binds themselves
	if key:match("ALT$") or key:match("SHIFT$") or key:match("CTRL$") or key=="UNKNOWN" then
		return -- these are already modifiers!
	end
	if dialog.purpose~="action" then -- if this isn't binding an action, and a modmod key itself
		if key=="MOUSEWHEELUP" or key=="MOUSEWHEELDOWN" then
			return -- this is mousewheel, not applicable to modmod key (no down+up event)
		end
		if key=="LeftButton" or key=="RightButton" then
		-- left and right buttons are so important to game function, disabling their use
			dialog.inset.message:SetText("The left and right mouse buttons can't be used as ModMod keys.")
			return
		end
		if IsShiftKeyDown() then
			key = "SHIFT-"..key
		end
		if IsControlKeyDown() then
			key = "CTRL-"..key
		end
		if IsAltKeyDown() then
			key = "ALT-"..key
		end
	end
	-- rename the primary 3 clicks as BUTTON1, BUTTON2 and BUTTON3
	if key=="LeftButton" then
		key = "BUTTON1"
	elseif key=="RightButton" then
		key = "BUTTON2"
	elseif key=="MiddleButton" then
		key = "BUTTON3"
	end
	key = key:upper() -- Button4+ will change to BUTTON4+ here

	if dialog.purpose=="action" then
		if key==GetBindingKey(format(CLICK_FORMAT,modmod.selectedModMod)) then
			dialog.inset.message:SetText("You can't bind an action to the ModMod key itself.")
			return
		end
	end

	modmod:DialogSetBind(key)
	local action = GetBindingAction(key)
	if dialog.purpose~="action" and action and action~="" then
		if action:match("ModModKey") then
			action = "another ModMod key"
		end
		-- if this is a new or edit of a modmod key, it's going to be an actual bind
		-- warn user that it will replace an existing one
		dialog.inset.message:SetText(format("\124cffffffff%s\124r is already bound to \124cffffffff%s\124r.\nOverwrite this binding?",GetBindingText(key),_G["BINDING_NAME_"..action] or action))
	end
end

-- a click can have multiple keys bound to it, this removes them all
function modmod:DeleteBindsForModModIndex(index)
	for _,key in ipairs({GetBindingKey(format(CLICK_FORMAT,index))}) do
		if key~="" then
			SetBinding(key,nil)
		end
	end
end

function modmod:DialogBindSave()
	PlaySound(PlaySoundKitID and "igMainMenuOptionCheckBoxOff" or SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_OFF)
	dialog:Hide()
	local binding = dialog.inset.bind.binding
	local purpose = dialog.purpose
	if purpose=="action" then
		-- first delete any key already assigned to the action
		for key,action in pairs(saved[modmod.selectedModMod]) do
			if action==dialog.action then
				saved[modmod.selectedModMod][key] = nil
			end
		end
		-- then save the new binding
		if binding then
			saved[modmod.selectedModMod][binding] = dialog.action
		end
	elseif purpose=="new" or purpose=="edit" then
		if purpose=="new" then
			tinsert(saved,{})
			modmod.selectedModMod = #saved
		end
		-- delete any existing binding first
		modmod:DeleteBindsForModModIndex(modmod.selectedModMod)
		-- and if there's a new binding, set it
		if binding then
			SetBinding(binding,format(CLICK_FORMAT,modmod.selectedModMod))
		end
		SaveBindings(GetCurrentBindingSet())
		-- if any actions use this key, delete it
		for key,action in pairs(saved[modmod.selectedModMod]) do
			if key==binding then
				saved[modmod.selectedModMod][key] = nil
			end
		end
	end
	modmod:UpdateSecureStuff()
	modmod:UpdateUI()
end

--[[ Secure Stuff ]]

-- these buttons are where the magic happens: the modmods are CLICKs to these
modmod.secureKeys = {}
for i=1,10 do
	modmod.secureKeys[i] = CreateFrame("Button","ModModKey"..i,nil,"SecureHandlerClickTemplate")
	modmod.secureKeys[i]:RegisterForClicks("AnyDown","AnyUp")
	modmod.secureKeys[i]:SetAttribute("_onclick",[[
		if down then -- if key goes down, we set override bindings
			for i=1,(self:GetAttribute("numBinds") or 0) do
				self:SetBinding(true,self:GetAttribute("bindingKey"..i),self:GetAttribute("bindingAction"..i))
			end
		else -- when key goes up, we remove override bindings
			self:ClearBindings()
		end
	]])
end

-- this goes through and sets bindings to the modmod keys and their actions
function modmod:UpdateSecureStuff()
	for i=1,#saved do
		local numBinds = 0
		local button = modmod.secureKeys[i]
		for key,action in pairs(saved[i]) do
			numBinds = numBinds + 1
			button:SetAttribute("bindingKey"..numBinds,key)
			button:SetAttribute("bindingAction"..numBinds,action)
		end
		button:SetAttribute("numBinds",numBinds)
	end
	for i=#saved+1,MAX_MODMODS do
		modmod.secureKeys[i]:SetAttribute("numBinds",0)
	end
end
